@@ -92,8 +92,8 @@ and fonts registered.
 package ODF::lpOD_Helper;
 
 { no strict 'refs'; ${__PACKAGE__."::VER"."SION"} = 997.999; }
-our $VERSION = '6.001'; # VERSION from Dist::Zilla::Plugin::OurPkgVersion
-our $DATE = '2023-09-02'; # DATE from Dist::Zilla::Plugin::OurDate
+our $VERSION = '6.002'; # VERSION from Dist::Zilla::Plugin::OurPkgVersion
+our $DATE = '2023-09-07'; # DATE from Dist::Zilla::Plugin::OurDate
 
 use Carp;
 use Data::Dumper::Interp 6.004 qw/visnew
@@ -113,7 +113,8 @@ our @EXPORT_OK = qw(
   hashtostring arraytostring
   AUTO_PFX
   Hr_MASK
-  TEXTLEAF_COND PARA_COND TEXTCONTAINER_COND TEXTLEAF_OR_PARA_COND
+  TEXTLEAF_FILTER PARA_FILTER TEXTCONTAINER_FILTER TEXTLEAF_OR_PARA_FILTER
+  ROW_FILTER COLUMN_FILTER CELL_FILTER TABLE_FILTER
 );
 
 use constant {
@@ -141,17 +142,14 @@ BEGIN {
 
 =head1 Transparent Unicode Support
 
-By default ODF::lpOD_Helper patches ODF::lpOD so that methods
+By default ODF::lpOD_Helper patches ODF::lpOD so that all methods
 accept and return arbitrary Perl character strings.
 
 You will B<always> want this unless
 your application really, really needs to pass un-decoded octets
 directly between file/network resources and ODF::lpOD without
-looking at the data along the way.
-Please see B<< L<ODF::lpOD_Helper::Unicode> >>.
-
-This can be disabled for legacy applications as described in
-L<ODF::lpOD_Helper::Unicode>.
+looking at the data along the way. This can be disabled for
+legacy applications.  Please see B<< L<ODF::lpOD_Helper::Unicode> >>.
 
 Currently this patch has global effect but might someday become
 scoped; to be safe put C<use ODF::lpOD_Helper> at the top of every file which
@@ -457,7 +455,7 @@ OPTIONS may be
   multi  => BOOL    # Allow multiple matches? (FALSE by default)
 
   prune_cond => STRING or qr/Regex/
-                    # Do not descend into nodes matching the indicated
+                    # Do not descend below nodes matching the indicated
                     # condition.  See "Hnext_elt".
 
 A hash is returned for each match:
@@ -510,8 +508,8 @@ In scalar context, a hashref or undef if there was no match
 
 =head2 B<Regex Anchoring>
 
-A qr/regex/ is matched against the combined virtual text of each paragraph.
-The match logic is
+If I<$expr> is a qr/regex/ it is matched against the combined
+virtual text of each paragraph.  The match logic is
 
    $paragraph_text =~ /\G.*?(${your_regex})/
 
@@ -566,7 +564,7 @@ with C<multi =E<gt> TRUE>, all instances are replaced.
 In the second form, the specified sub is called for each match, passing
 a I<match hashref> (see C<Hsearch>) as the only argument.  Its return
 value determines whether any substitutions occur.
-The sub must return one of the following:
+The sub must
 
  return(0)
 
@@ -578,20 +576,15 @@ The sub must return one of the following:
     starting immediately after the replaced text.
 
  return(Hr_SUBST | Hr_STOP, [content])
- return(Hr_SUBST | Hr_STOP, [content], optRESULTS)
 
     [content] is substituted for the matched text and then "Hreplace"
     terminates immediately.
 
-    If optRESULTS is provided, it is returned from "Hreplace" instead
-    of the default substitution-descriptor hashes.
-
  return(Hr_STOP)
- return(Hr_STOP, optRESULTS)
 
     "Hreplace" just terminates.
 
-C<Hreplace> returns, by default, a list of hashes describing the
+C<Hreplace> returns a list of hashes describing the
 substitutions which were performed:
 
   {
@@ -702,20 +695,25 @@ sub Hor_cond(@) {
   oops avis(@_);
 }#Hor_cond
 
-use constant TEXTLEAF_COND         => '#TEXT|text:tab|text:line-break|text:s';
-use constant PARA_COND             => 'text:p|text:h';
-use constant TEXTCONTAINER_COND    => Hor_cond(PARA_COND, "text:span");
-use constant TEXTLEAF_OR_PARA_COND => Hor_cond(TEXTLEAF_COND, PARA_COND);
+use constant ROW_FILTER              => ODF::lpOD::Matrix::ROW_FILTER;    # 'table:table-row'
+use constant COLUMN_FILTER           => ODF::lpOD::Matrix::COLUMN_FILTER; # 'table:table-column'
+use constant CELL_FILTER             => ODF::lpOD::Matrix::CELL_FILTER;   # qr'table:(covered-|)table-cell'
+use constant TABLE_FILTER            => ODF::lpOD::Matrix::TABLE_FILTER;  # 'table:table'
+
+use constant TEXTLEAF_FILTER         => '#TEXT|text:tab|text:line-break|text:s';
+use constant PARA_FILTER             => 'text:p|text:h';
+use constant TEXTCONTAINER_FILTER    => Hor_cond(PARA_FILTER, "text:span");
+use constant TEXTLEAF_OR_PARA_FILTER => Hor_cond(TEXTLEAF_FILTER, PARA_FILTER);
 
 # These used to be lexical subs inside Hreplace, but a Perl bug prevented
-# reentrant searches (via callbacks).   The stuff in the $state hash
+# reentrant searches via callbacks.   The stuff in the $state hash
 # used to be lexicals in Hreplace.  https://github.com/Perl/perl5/issues/18606
   sub _get_seginfo($$) {
     my ($state, $node) = @_;
     my @seginfo;
     my $vtext = "";
     # Do not descend into nested paragraphs, which are visited in outer loop
-    for my $e ($node->Hdescendants_or_self(TEXTLEAF_COND, PARA_COND)) {
+    for my $e ($node->Hdescendants_or_self(TEXTLEAF_FILTER, PARA_FILTER)) {
       my $etext = __leaf2vtext($e);
       my $textlen = length($etext);
       push @seginfo, {
@@ -801,6 +799,7 @@ use constant TEXTLEAF_OR_PARA_COND => Hor_cond(TEXTLEAF_COND, PARA_COND);
   }# _first_match
 
   sub _process_para($$) { # also called if context is a text leaf
+    # returns 0 to continue, Hr_STOP to stop
     my ($state, $node) = @_; # paragraph or total $context if a textual leaf
     my ($debug) = @$state{qw/debug/};
 
@@ -825,6 +824,8 @@ use constant TEXTLEAF_OR_PARA_COND => Hor_cond(TEXTLEAF_COND, PARA_COND);
         if (defined $m) {
 oops unless @{$m->{segments}};
           if ($r & Hr_SUBST) {
+            croak ivis 'Extraneous return values from callback @args'
+              if @args > 1;
             my $content = shift @args // confess "No [content] after Hr_SUBST";
             my $new_vlength = Hreplace_match($m, $content, debug => $debug);
 
@@ -834,7 +835,7 @@ oops unless @{$m->{segments}};
               vlength      => $new_vlength,
               para         => $m->{para},
               para_voffset => $m->{para_voffset},
-            } ;#if defined wantarray;
+            } ;#if defined wantarray in Hreplace();
 
             # Re-process the whole paragraph with $offset set appropriately.
             $state->{offset} = $m->{voffset}; # start of match
@@ -854,6 +855,8 @@ oops unless @{$m->{segments}};
               redo PARA;
             }
           } else {
+            croak ivis 'Extraneous return values from callback @args'
+              if @args > 0;
             if ($m->{match} eq "") {
               # Avoid infinite loop
               btw dvis '  Hr NULL MATCH: offset++' if $debug;
@@ -864,15 +867,14 @@ oops unless @{$m->{segments}};
           }
           if ($r & Hr_STOP) {
             $node->Hnormalize() if $para_startcount != scalar @{$state->{subst_results}};
-            #btw ivis '  Hr STOP  retvals=@args' if $debug; #logged in mainbody
-            return(Hr_STOP, @args);
+            return(Hr_STOP)
           }
           if ($state->{offset} < $state->{totlen}) {
             btw dvis '  Hr CONTINUE: new $state->{offset}, *redo MATCH*' if $debug;
             redo MATCH;
           }
         } else {
-          btw '  Hr [no match] expr=',vis($state->{expr}),dvis '$state $vtext' if $debug;
+          btw '  Hr [no match] expr=',vis($state->{expr}),dvis '\n $state $vtext' if $debug;
         }
       }#MATCH
     }#PARA
@@ -897,7 +899,7 @@ sub ODF::lpOD::Element::Hreplace {
   #
   # Attempted modifications throw an exception if $recursion_level > 1
   # i.e. only the outer-most level is allowed to change anything.
-  # This is checked in Hreplace_match (in case we ever make it public)
+  # This is checked in Hreplace_match (in case we ever make that public)
   local $recursion_level = $recursion_level + 1;
 
   btw dvis 'Hreplace Top: $context $expr $repl %opts' if $debug;
@@ -961,28 +963,24 @@ sub ODF::lpOD::Element::Hreplace {
 
   my ($stop, @retvals);
   # If $context itself is a paragraph or a leaf segment, process it first
-  if ($context->passes(TEXTLEAF_OR_PARA_COND)) {
-    ($stop, @retvals) = _process_para($state, $context); # ignores nested paras
+  if ($context->passes(TEXTLEAF_OR_PARA_FILTER)) {
+    $stop = _process_para($state, $context); # ignores nested paras
   }
   # Now process paragraphs which are descendants.
   # _process_para will not descend into nested paragraphs, i.e. it only
   # looks at the text at that level; nested paragraphs are visited directly
   # here (after the parent paragraph was visited).
   { my $para = $context ;
-    while ($para = $para->Hnext_elt($context, PARA_COND, $opts{prune_cond})) {
-      ($stop, @retvals) = _process_para($state, $para);
+    while ($para = $para->Hnext_elt($context, PARA_FILTER, $opts{prune_cond})) {
+      $stop = _process_para($state, $para);
       last if $stop & Hr_STOP;
     }
   }
 
   if ($stop & Hr_STOP) {
     btw dvis 'Hreplace STOP. @retvals' if $debug;
-    croak ("Callback specified results to return but context is void")
-      if @retvals && !defined(wantarray);
-    croak ("Callback specified multiple results but context is scalar")
-      if @retvals > 1 && !wantarray;
   }
-  @retvals = @{$state->{subst_results}} if @retvals == 0;
+  @retvals = @{$state->{subst_results}};
 
   if (wantarray) {
     btw ivis 'Hreplace RETURNING @retvals' if $debug;
@@ -1132,23 +1130,23 @@ sub ODF::lpOD::Element::Hoffset_into_vtext {
   my ($context, $offset, $prune_cond) = @_;
   confess ivis 'Invalid offset $offset' if ($offset//-1) < 0;
   my $remaining = $offset;
-  my $elt = $context->passes(TEXTLEAF_COND)
-              // $context->Hnext_elt($context, TEXTLEAF_COND, $prune_cond);
+  my $elt = $context->passes(TEXTLEAF_FILTER)
+              // $context->Hnext_elt($context, TEXTLEAF_FILTER, $prune_cond);
   my $last_leaf;
   while ($elt) {
     my $text = __leaf2vtext($elt);
     if ($remaining >= length($text)) {
       $remaining -= length($text);
       $last_leaf = $elt;
-      $elt = $elt->Hnext_elt($context, TEXTLEAF_COND, $prune_cond);
+      $elt = $elt->Hnext_elt($context, TEXTLEAF_FILTER, $prune_cond);
       next
     }
     return ($elt, $remaining);
-#    my $para = $elt->parent(PARA_COND);
+#    my $para = $elt->parent(PARA_FILTER);
 #    my $poffset = 0;
 #    my $ptxt = $para;
 #    for(;;) {
-#      $ptxt = $ptxt->Hnext_elt($para,TEXTLEAF_COND,$prune_cond) // oops;
+#      $ptxt = $ptxt->Hnext_elt($para,TEXTLEAF_FILTER,$prune_cond) // oops;
 #      last if $ptxt == $elt;
 #      $poffset += length(__leaf2vtext($ptxt));
 #    }
@@ -1247,9 +1245,9 @@ sub ODF::lpOD::Element::Hinsert_element {
     }
     # There are _no_ textual leaves.  Insert as FIRST_CHILD of container
     oops unless $offset==0;
-    oops if $context->passes(TEXTLEAF_COND);
-    my $container = $context->passes(TEXTCONTAINER_COND)
-                         ? $context : $context->next_elt(TEXTCONTAINER_COND);
+    oops if $context->passes(TEXTLEAF_FILTER);
+    my $container = $context->passes(TEXTCONTAINER_FILTER)
+                         ? $context : $context->next_elt(TEXTCONTAINER_FILTER);
     confess "context is not a/has no text container" unless $container;
     return $container->insert_element($to_insert, position => FIRST_CHILD)
   }
@@ -1356,7 +1354,7 @@ sub ODF::lpOD::Element::Hinsert_content($$) {
   my %opts = (position => FIRST_CHILD, @_);
   my $debug = $opts{debug};
 
-  my sub show_context {
+  my sub _show_context {
     my $msg = join("", @_);
     $msg .= "context="._abbrev_addrvis($context);
     my $item = $context;
@@ -1365,8 +1363,7 @@ sub ODF::lpOD::Element::Hinsert_content($$) {
       $msg .= " (showing ancestor para)";
     }
     $msg .= " :\n".fmt_tree($item);
-    @_ = ($msg);
-    goto &btw;  # show caller's line number
+    btwN 1,$msg;
   }
 
   # Hinsert_content might become a superset (i.e. drop-in replacement
@@ -1381,7 +1378,7 @@ sub ODF::lpOD::Element::Hinsert_content($$) {
 
   my @content = @$what;
 
-  show_context(dvis '##Hi_c TOP %opts\n @content\n') if $debug;
+  _show_context(dvis '##Hi_c TOP %opts\n @content\n') if $debug;
 
   croak "option 'Chomp' was renamed 'chomp'" if exists $opts{Chomp};
   if ($opts{chomp} && @content) {
@@ -1490,7 +1487,7 @@ sub ODF::lpOD::Element::Hinsert_content($$) {
   _cleanup_spans(\@nodes, %opts) unless $opts{_nocleanup};
 
   my $result = { vlength => $vlength // 0 };
-  show_context(dvis '##Hi_c RESULT: $result\n') if $opts{debug};
+  _show_context(dvis '##Hi_c RESULT: $result\n') if $opts{debug};
 
   confess "Hinsert_content() returns a scalar (a hashref)\n" if wantarray;
   $result;
@@ -1525,7 +1522,7 @@ and the original node will be empty upon return.
 if $offset equals the existing length then the new sibling will be empty.
 
 If a text:s node is split then the new node will also be a text:s node
-"containing" the appropriate number of spaces.  The 'c' attribute will
+"containing" the appropriate number of spaces.  The 'text:c' attribute will
 be zero if the node is "empty".
 
 If a text:tab or text:line-break node is split then either the new node
@@ -1536,11 +1533,11 @@ to become an empty PCDATA node.
 
 sub ODF::lpOD::Element::His_textual {
   my $elt = shift;
-  $elt->passes(TEXTLEAF_COND)
+  $elt->passes(TEXTLEAF_FILTER)
 }
 sub ODF::lpOD::Element::His_text_container {
   my $elt = shift;
-  $elt->passes(TEXTCONTAINER_COND)
+  $elt->passes(TEXTCONTAINER_FILTER)
 }
 
 sub ODF::lpOD::Element::Hsplit_element_at {
@@ -1548,7 +1545,7 @@ sub ODF::lpOD::Element::Hsplit_element_at {
   confess "Wrong number of arguments" unless @_==2;
   # see XML::Twig::split_at
   my $text_elt= $elt->His_textual() ? $elt
-                                    : $elt->first_child(TEXTLEAF_COND)
+                                    : $elt->first_child(TEXTLEAF_FILTER)
                                         || confess("no textual leaf found");
   if ($text_elt->tag eq 'text:s') {
     my $existing_count = $text_elt->get_attribute('c') // 1;
@@ -1568,7 +1565,7 @@ sub ODF::lpOD::Element::Hsplit_element_at {
     return $text_elt->split_at($offset);
   }
   elsif ($offset == 0) {
-#my $para = $elt->passes(PARA_COND) ? $elt : $elt->parent(PARA_COND);
+#my $para = $elt->passes(PARA_FILTER) ? $elt : $elt->parent(PARA_FILTER);
 #btw dvis 'BEFORE TRANSMUTE $para = ',fmt_tree($para);
 #btw dvis 'BEFORE TRANSMUTE $elt = ',fmt_tree($elt);
     my $result =
@@ -1605,7 +1602,7 @@ sub ODF::lpOD::Element::Hsplit_element_at {
 =head2 $context->Hget_text(prune_cond => COND)
 
 Gets the combined "virtual text" in or below C<$context>,
-including in any nested paragraphs (e.g. in Frames or Tables).
+by default including in nested paragraphs (e.g. in Frames or Tables).
 The special nodes which represent tabs, line-breaks and
 consecutive spaces are expanded to the corresponding characters.
 
@@ -1637,9 +1634,8 @@ I think get_text's "recursive" option was probably intended to include text from
 paragraphs in possibly-nested frames and tables, and it was an oversight
 that that special text nodes are not always handled correctly.
 
-Note that there is no 'recursive' option to C<Hget_text>, which behaves
-that way by default.  Hget_text offers the 'prune_cond' option to
-restrict expansion.
+Note that B<Hget_text> is "recursive" by default; the 'prune_cond'
+option is the only way to restrict recursion.
 
 =cut
 
@@ -1651,7 +1647,7 @@ sub ODF::lpOD::Element::Hget_text {
   # by ODF::lpOD, and now does nothing unless our ':bytes' tag is imported.
   ODF::lpOD::Common::output_conversion(
     join "", map{ __leaf2vtext($_) }
-             $self->Hdescendants_or_self(TEXTLEAF_COND, $opt{prune_cond})
+             $self->Hdescendants_or_self(TEXTLEAF_FILTER, $opt{prune_cond})
   )
 }
 
@@ -1809,7 +1805,7 @@ work exactly like the correspoinding non-B<H> methods.
 
 C<Hnext_elt>, C<Hdescendants> and C<Hdescendants_or_self>
 C<Hparent> and C<Hself_or_parent>
-are installed as methods of XML::Twig::Elt.
+are installed as methods of I<XML::Twig::Elt>.
 
 =cut
 
@@ -1868,20 +1864,16 @@ sub XML::Twig::Elt::Hdescendants_or_self {
 
 Returns the nearest ancestor which matches condition C<$cond>.
 
-If C<$stop_cond> is defined, then undef is returned if the search would
+If C<$stop_cond> is defined, then 0 is returned if the search would
 ascend above the nearest ancestor matching the stop condition.
-An exception is thrown if no ancestor matches either $cond or $stop_cond.
+Undef is returned no ancestor matches either $cond or $stop_cond.
 
 For exmaple,
 
   my $row = $elt->Hparent("table:table-row", "draw:frame");
 
-would locate the table row containing $elt but return undef if $elt was
-encapsulated in a frame within a row.
-
-If you want to avoid an exception if '$cond' is not found then you
-can include 'office:text' in C<$stop_cond>, which stops at the
-root of the document body.
+would locate the table row containing $elt but return false if $elt was
+encapsulated in a frame within the row (or if it is in a table at all).
 
 =head2 $node->Hself_or_parent($cond, [$stop_cond])
 
@@ -1892,9 +1884,9 @@ Like C<Hparent> but returns $node itself if it matches $cond.
 sub XML::Twig::Elt::Hparent($$;$) {
   my ($elt, $cond, $stop_cond) = @_;
   for(;;) {
-    $elt = $elt->{parent} || croak "No ancestors match ",vis($cond);
+    $elt = $elt->{parent} || return undef;
     return $elt if $elt->passes($cond);
-    return undef if $stop_cond && $elt->passes($stop_cond);
+    return 0 if $stop_cond && $elt->passes($stop_cond);
   }
 }
 
@@ -1910,18 +1902,18 @@ sub XML::Twig::Elt::Hself_or_parent($$;$) {
 
 This function combines multiple XML::Twig search conditions,
 which may be any mixture of string, regex, or code-ref conditions.
-The resulting condition will match any of the input conditions ("or").
+The resulting condition will match any of the input conditions (hence "or").
 
 This is useful to augment conditions exported by another module
 when you are not certain how the other condition is implemented, for example
 
-  use ODF::lpOD_Helper qw(:DEFAULT PARA_COND);
-  use constant MY_PARAORFRAME_COND => Hor_cond(PARA_COND, 'draw:frame');
+  use ODF::lpOD_Helper qw(:DEFAULT PARA_FILTER);
+  use constant MY_PARAORFRAME_FILTER => Hor_cond(PARA_FILTER, 'draw:frame');
   ...
-  @elts = $context->descendants(MY_PARAORFRAME_COND)
+  @elts = $context->descendants(MY_PARAORFRAME_FILTER)
 
 This would collect all paragraphs or frames below $context.
-Note that PARA_COND might be 'text:p|text:h' or qr/^text:[ph]$/
+Note that PARA_FILTER might be 'text:p|text:h' or qr/^text:[ph]$/
 or C<sub{ $_[0] eq 'text:p' || $_[0] eq 'text:h' }> etc.
 
 C<Hor_cond> optimizes a few regex forms into equivalent string conditions,
@@ -1941,7 +1933,7 @@ In the case of a I<style>, the C<$family> must be specified
 ("text", "table", etc.).
 
 SUFFIX is an optional string which will be appended to a generated
-unique name (to make it easier for humans to recognize).
+unique name to make it easier for humans to recognize.
 
 C<$context> may be the document itself or any Element.
 
@@ -2139,14 +2131,17 @@ sub ODF::lpOD::Document::Hcommon_style($$@) {
 
 =head2 arraytostring($arrayref)
 
-Returns a string uniquely representing the datum (hash or array).
-Any references within the datum are represented using their 'refaddr'
-i.e. the result will not match a datum with different sub-elements
-even if they have the same final content.
+Returns a signature string uniquely representing the members
+(keys and values in the case of a hash).  References are represented
+using their 'refaddr'.
+
+Note that referenced data is not recursively examined.
+Signatures of different structures will match only if
+corresponding first-level non-ref values are 'eq' and refs are exactly the same refs.
 
 =cut
 
-sub _item_rep(_) {
+sub _item_rep(_) {  # encode so that a string can not fake something else
   ref($_[0]) ? "~".refaddr($_[0])        :
   substr($_[0],0,1) eq "~" ? visq($_[0]) :
   index($_[0],"!") >= 0 ? vis($_[0])     :
@@ -2166,18 +2161,19 @@ sub arraytostring($) {
 
 ###################################################
 
-=head2 fmt_node($node)
+=head2 fmt_node($node, OPTIONS)
 
-Format a single node for debug messages, without a final newline.
+Format a single ODF::lpOD (really XML::Twig) node for
+debug messages, without a final newline.
 
-C<wi =E<gt> NUM> may be given as assitional arguments
+C<wi =E<gt> NUM> may be given in OPTIONS
 to indent wrapped lines by the indicated number of spaces.
 
-=head2 fmt_tree($subtree_root)
+=head2 fmt_tree($subtree_root, OPTIONS)
 
 Format a node and all of it's children (sans final newline).
 
-=for Pod::Coverage TEXTLEAF_COND PARA_COND TEXTCONTAINER_COND
+=for Pod::Coverage TEXTLEAF_FILTER PARA_FILTER TEXTCONTAINER_FILTER
 =for Pod::Coverage fmt_node_brief fmt_tree_brief fmt_match
 =cut
 
@@ -2207,7 +2203,7 @@ sub _abbrev_addrvis($) {  # like DDI::addrvis() but elides ODF::lpOD:: prefix
 #
 sub fmt_node(_;@) {  # sans final newline
   my $node = shift;
-  my %opts = (showaddr => TRUE, showlen => TRUE, showisa => TRUE, @_);
+  my %opts = (showlevel => FALSE, showaddr => TRUE, showlen => TRUE, showisa => TRUE, @_);
   return "undef" unless defined($node);
   oops unless ref($node);
   $opts{wi} //= 0;  # wi = "wrap indent"
@@ -2242,13 +2238,15 @@ sub fmt_node(_;@) {  # sans final newline
     $tagopen .= " ";
   }
   $tagopen .= "{";
-  if (defined $opts{_level}) {
+  if ($opts{showlevel}) {
     $tagopen .= $opts{_level};
     ++$opts{_level};
   }
 
   my @middle_parts;
-  my $mid_visobj = visnew->Pad($wrapspace.$indent_incr_space);
+  my $mid_foldwidth = visnew->Foldwidth() - (($opts{wi}//0) + $indent_incr);
+  my $mid_visobj = visnew->Pad($wrapspace.$indent_incr_space)
+                         ->Foldwidth($mid_foldwidth);
 
   if (defined($opts{parent})) {
     push @middle_parts,
@@ -2260,10 +2258,15 @@ sub fmt_node(_;@) {  # sans final newline
       if $opts{showaddr};
   }
   if (defined(my $att  = $node->get_attributes)) {
-    if (($tag//"") =~ /^(table-cell|sequence)/) {
+    if (($tag//"") =~ /^(table-cell|sequence)/ && !$opts{showall}) {
       push @middle_parts, "att={...}"; # voluminous & uninteresting
     } else {
-      push @middle_parts, "att=".$mid_visobj->vis($att);
+      my %edited_att = map{ ($_ => $att->{$_}) }
+                       grep{ !/^#lpod:/ }  # e.g. #lpod:part circular reference
+                       keys %$att;
+      push @middle_parts,
+           #"att=".$mid_visobj->visq(\%edited_att);
+           "att=".$mid_visobj->Foldwidth1($mid_foldwidth-4)->visq(\%edited_att);
     }
   }
 
@@ -2322,7 +2325,7 @@ sub fmt_node(_;@) {  # sans final newline
   }
 
   my $tagclose = "}";
-  if (defined $opts{_level}) {
+  if ($opts{showlevel}) {
     --$opts{_level};
     $tagclose .= $opts{_level};
   }
@@ -2360,8 +2363,8 @@ sub fmt_node_brief(_;@) {
 
 sub fmt_tree(_;@) { # sans final newline
   my $top = shift;
-  my %opts = (showoff => TRUE, @_);
-  my $indent = $opts{indent};
+  my %opts = (showoff => TRUE, showlevel => TRUE, @_);
+  my $indent = $opts{indent} // 0;
   my $string = "";
   my $parent;
   if ($opts{ancestors} and ref $top) {
@@ -2378,10 +2381,7 @@ sub fmt_tree(_;@) { # sans final newline
   my $vtoffset = 0;
   $opts{_vtoref} = \$vtoffset if $opts{showoff};
   $opts{_level} = 0;
-  $string .= fmt_node($top, %opts, _leaftextonly => 1, _recursive => 1);
-
-  #return "------------\n".$string."------------";
-  return $string;
+  fmt_node($top, %opts, _leaftextonly => 1, _recursive => 1);
 }
 sub fmt_tree_brief(_;@) {
   fmt_tree($_[0], showaddr=>FALSE, showlen=>FALSE, showoff=>FALSE, @_[1..$#_])
@@ -2460,33 +2460,35 @@ using "rsid" styles which interfere with cloning.
 The problem is that LO expectes these styles to be referenced exactly
 once.  The C<Hclean_for_cloning()> method will remove them.
 
-It may also be possible to save LibreOffice documents without 'rsids' :
+There is some evidence that it was supposed to be possible to save LO documents
+without 'rsid's but AFAIK such a feature is not currently available. See
 
-  https://ask.libreoffice.org/t/where-do-the-text-style-name-tnn-span-tags-come-from-and-how-do-i-get-rid-of-them/31681/2
+=over
 
-  https://bugs.documentfoundation.org/show_bug.cgi?id=68183
+L<https://ask.libreoffice.org/t/where-do-the-text-style-name-tnn-span-tags-come-from-and-how-do-i-get-rid-of-them/31681/2>
+
+L<https://bugs.documentfoundation.org/show_bug.cgi?id=68183>
+
+=back
 
 =head2 $doc->Hclean_for_cloning();
 
-This unpleasant hack removes any "rsid" styles.
+This unpleasant hack removes all "rsid" properties from all styles the document.
 
-C<Hclean_for_cloning> should be called before cloning any content
-in the document, if the cloned items might have been edited by Libre Office.
+C<Hclean_for_cloning> should be called before cloning anything
+in a document if the cloned items might have been edited by Libre Office.
 It may be called multiple times; second and subsequent calls do nothing.
 
 Gory detail:
-Every span in the document body is examined;
-if it references a text style with
-a B<officeooo:rsid> or B<officeooo:paragraph-rsid>
-attribute in a
-descendant style:text-properties node, then that attribute is removed.
-If the text-properties contains other attributes then everything
-else is left as-is (this is the case when the style has an additional
-purpose besides holding an rsid attribute).
-If the text-properties node has no other
-attributes it is deleted, and if the ancestor style has no surviving
-text-properties then the style is deleted and span(s) which
-reference it are erased, moving up the span's childen.
+Every style in the document is examined and
+B<officeooo:rsid> and B<officeooo:paragraph-rsid>
+attributes are deleted.
+
+(This next step is currently disabled:)
+Then every span in the document body is examined and if the span's
+style is a style which no longer has any properties (i.e. it existed
+only to record an rsid property), then the span is erased, moving up the
+span's children.
 
 =cut
 
@@ -2494,89 +2496,80 @@ reference it are erased, moving up the span's childen.
 # which if cloned or otherwise used for multiple spans cause Libre Office
 # to hang or crash.
 #
-# Modified styles which have no remaining attributes are deleted, and spans
-# which were useing them are erased (moving their content up one level).
 sub ODF::lpOD::Document::Hclean_for_cloning {
   my ($doc, %opts) = @_;
 
   my $sh = _get_ephemeral_statehash($doc);
   return if $sh->{cleaned_for_cloning}++;
 
-  # Collect the *names* of styles bearing rsid properties.
-  #  values are [$style_object, $property_objects....]
-  my %rsid_styles;
+  my %rsid_styles; # stylename => style
   {
-    # [OLD] This does not work due to a bug overloading "text" with the
-    # same-named DataStyle type:
-    # my @textstyles = $doc->get_styles('text');
-    #FIXME ... I can't see why get_styles('text') is not correct!
-    ##    DataStyle::is_numeric_family('text') should return false...
+    # Find all children of style:style
+    # (namely style:{paragraph,text}-properties) containing an rsid property
+    # and delete the property from the style.
 
-    # The following was extracted from ODF::lpOD::Document::get_styles
-    my $xp = '//style:style[@style:family="text"]';
-    my @textstyles = ($doc->get_elements(STYLES,$xp), $doc->get_elements(CONTENT,$xp));
-    foreach my $tstyle (@textstyles) {
-      my @props = $tstyle->descendants(qr/style:text-properties/);
-      if (any{
-            # Look for 'officeooo:rsid', 'officeooo:paragraph-rsid'
-            any{ /officeooo:.*rsid/ } keys(%{ $_->get_attributes })
-          } @props)
-      {
-        $rsid_styles{$tstyle->get_name} = [$tstyle, @props];
-      }
+    my $xp = '//style:style/[@officeooo:paragraph-rsid or @officeooo:rsid]';
+    my @items=($doc->get_elements(STYLES,$xp), $doc->get_elements(CONTENT,$xp));
+    foreach my $item (@items) {
+      # style:paragraph-properties or style:text-properties
+      my $atts = $item->atts;  # XML::Twig
+      # DELETE the rsid property
+      my @attnames = grep{ /rsid/ } keys %$atts;
+      oops unless @attnames;
+      delete @$atts{@attnames};
+      $item->set_atts($atts); # XML::Twig
+      my $style = $item->parent('style:style') // oops;
+      $rsid_styles{$style->get_name} = $style;
     }
+
+    my @items2=($doc->get_elements(STYLES,$xp), $doc->get_elements(CONTENT,$xp));
+    oops if @items2;
   }
 
+  # That should be enough; the body will still be littered with spans which
+  # use zero-effect styles (now that the rsid properties are gone).
+
+##  # For now, leave them be.
+#btw "Temp skipping rsid-only span erasure";
+##  return
+##    unless $ENV{ERASE_RSID_SPANS};
+##btw "Erasing rsid-only spans ...";
+
   # Check every span in the document body
   my $body = $doc->get_body;
   my $span = $body;
+  my @to_be_erased;
   while ($span = $span->next_elt($body, 'text:span')) {
     my $tsname = $span->get_attribute('style') // oops;
-    next unless (my $v = $rsid_styles{$tsname});
-    if (ref $v) { # not (yet) deleted
-      # Remove rsid properties from the style.  If other properties remain
-      # then the cleaned style is left; if there are other spans that
-      # reference the same style (which I don't think LibreOffice allows,
-      # but being defensive here...) then the style will be harmlessly
-      # re-examined again.
-      #
-      # If no properties remain after removing rsid properties, then the style
-      # is deleted and the value in %rsid_styles set to "DELETED", which
-      # signals that all spans referencing the style should be erased.
-      my ($tstyle, @props) = @$v;
-      oops unless $tstyle->{parent};
+    my $style = $rsid_styles{$tsname};
+    unless ($style) {
+#btw dvis 'MM0 $tsname not in rsid_styles' if $opts{debug};
+      next;
+    }
+#btw dvis 'MM1 $span $tsname $style' if $opts{debug};
+    if (ref $style) { # not (yet) deleted
+      oops unless $style->{parent};
       my $props_remain;
-      for my $text_prop (@props) {
-        oops unless $text_prop->{parent};
-        my %att = $text_prop->get_attributes;
-        foreach my $key (grep { /officeooo:.*rsid/ } keys %att) {
-          delete $att{$key}
-        }
-        if (keys %att) {
-          $text_prop->del_attributes;
-          $text_prop->set_attributes(%att);
-          $props_remain = 1;
-          btw "$tsname : now-rsid-free text_prop ",fmt_node($text_prop) if $opts{debug};
-        } else {
-          btw "$tsname : Deleting rsid-only text_prop ",fmt_node($text_prop) if $opts{debug};
-          $text_prop->delete;
-        }
+      foreach my $ch ($style->children) {
+         oops unless $ch->tag() =~ /^style:.*properties$/;
+         #my $atts = $item->atts;  # XML::Twig
+         my $atts = $ch->get_attributes;
+         ($props_remain=1,last) if keys %$atts;
       }
       if (! $props_remain) {
-        btw "$tsname : Deleting rsid style ",fmt_tree($tstyle, internals=>1)
+        btw "$tsname : Deleting rsid-only style ",fmt_tree($style, internals=>1)
           if $opts{debug};
-        $tstyle->delete;
-        $rsid_styles{$tsname} = $v = "DELETED";
+        $style->delete;
+        $rsid_styles{$tsname} = $style = "DELETED";
+      } else {
+        btw "$tsname : KEEPING style ",fmt_tree($style, internals=>1) if $opts{debug};
       }
     }
-    if ($v eq "DELETED") {
-      btw "Erasing rsid-only span ",fmt_node($span) if $opts{debug};
-      $span->erase; # XML::Twig
-      # FIXME: There were old comments in the code copied from RFFMDirGenODF
-      # to the effect that ODF::lpOD overrides XML::Twig methods with
-      # incompatible APIs and breaks XML::Twig::Elt::erase().
-      # But I don't see that now...
-    }
+    push @to_be_erased, $span if $style eq "DELETED";
+  }
+  foreach my $span (@to_be_erased) {
+    btw "Erasing rsid-only span ",fmt_node($span) if $opts{debug};
+    $span->erase; # XML::Twig
   }
 }#Hclean_for_cloning
 
@@ -2591,9 +2584,7 @@ As of Feb 2023, the underlying ODF::lpOD is not actively maintained
 However with ODF::lpOD_Helper, ODF::lpOD is once again an
 extremely useful tool.
 
-B<Original Motivation:>
-
-ODF::lpOD by itself can be inconvenient because
+B<Motivation:> ODF::lpOD by itself can be inconvenient because
 
 =over
 
@@ -2614,11 +2605,35 @@ I<replace()> has analogous limitations.
 
 =item 3.
 
+Nested paragraphs (which can occur via frames and tables) are difficult
+or impossible to deal with because of various limitations of L<ODF::lpOD>
+(notably with C<get_text()>).
+
+=item 4.
+
 "Unknown method DESTROY" warnings occur without a patch (ODF::lpOD v1.126;
 L<https://rt.cpan.org/Public/Bug/Display.html?id=97977>)
 
 =back
 
+=head2 Why not just fix ODF::lpOD ?
+
+Ideally bugs in L<ODF::lpOD> would simply be fixed.
+Enhancements should be added in a compatible way
+with integrated documentation for everything.
+
+However the original author of ODF::lpOD, Jean-Marie Gouarne,
+is no longer active and some bugs (notably with C<get_text>) appear
+to require non-trivial changes.
+Making substantial changes to the ODF::lpOD code base
+would be too risky at this point in history.
+
+ODF::lpOD_Helper introduced higher-level APIs which should
+get further study and revision to fully comport with Jean-Marie's
+vision and usage model.
+Ideally that would be done but it would require too much effort
+(again, at this point in history).  At least for this author!
+
 =head1 AUTHOR
 
 Jim Avera  (jim.avera AT gmail)
